################################################################################
#  Copyright (C) 2006  Travis Shirk <travis@pobox.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  $Id$
################################################################################
import gtk, gtk.glade

import mesk
from mesk.i18n import _
from mesk.plugin import PluginMgr
import mesk.window

from dialogs import ErrorDialog

class PreferenceWindow(mesk.window.Window):
    def __init__(self):
        mesk.window.Window.__init__(self, 'preference_window',
                                    'preference_window.glade')

        self.prefs_notebook = self.xml.get_widget('pref_notebook')
        self.close_button = self.xml.get_widget('close_button')
        self.plugins_notebook = self.xml.get_widget('plugins_notebook')
        self.plugins_notebook.set_show_tabs(False)
        self.plugin_config_container = self.xml.get_widget('plugin_config_vbox')

        # Plugin info widgets
        self._name_label = self.xml.get_widget('name_label')
        self._desc_label = self.xml.get_widget('description_label')
        self._author_label = self.xml.get_widget('author_label')
        self._copyright_label = self.xml.get_widget('copyright_label')
        self._plugin_image = self.xml.get_widget('plugin_image')
        self._config_button = self.xml.get_widget('plugin_config_button')
        self._config_button.set_sensitive(False)

        self._url_linkbutton = gtk.LinkButton('')
        from mesk.gtk_utils import default_linkbutton_callback
        self._url_linkbutton.connect('clicked', default_linkbutton_callback)

        # Create plugins list and model
        (self.MODEL_ENABLED,
         self.MODEL_NAME) = range(2)
        plugins_model = gtk.ListStore(int, # Enabled
                                      str, # Plugin name
                                     )

        plugins_view = self.xml.get_widget('plugins_treeview')
        plugins_view.set_model(plugins_model)

        col = gtk.TreeViewColumn(_('Plugin'))
        txt_renderer = gtk.CellRendererText()
        col.pack_start(txt_renderer, True)
        col.add_attribute(txt_renderer, 'text', 1)
        plugins_view.append_column(col)

        col = gtk.TreeViewColumn(_('Enabled'))
        toggle_renderer = gtk.CellRendererToggle()
        toggle_renderer.set_property('activatable', True)
        toggle_renderer.connect('toggled', self._on_plugin_toggled,
                                plugins_model)
        col.pack_start(toggle_renderer, True)
        col.add_attribute(toggle_renderer, 'active', 0)
        plugins_view.append_column(col)

        # Populate plugins list
        self._plugins = list(mesk.plugin.get_manager().get_all_plugins())
        self._plugins.sort(cmp=lambda x,y: cmp(x['DISPLAY_NAME'],
                                               y['DISPLAY_NAME']))
        active_plugins = mesk.plugin.get_manager().get_active_plugins()
        if self._plugins:
            active_names = []
            for plugin in active_plugins:
                active_names.append(plugin.name)

            for plugin in self._plugins:
                plugins_model.append([plugin['NAME'] in active_names,
                                      plugin['DISPLAY_NAME']])
                # Create a Pixbuf from the xpm data
                pixbuf = gtk.gdk.pixbuf_new_from_xpm_data(plugin['XPM'])
                plugin['PIXBUF'] = pixbuf
                # Save some memory
                plugin['XPM'] = ''

            plugins_view.set_cursor(0)
        else:
            # TODO: Hide plugins tab
            pass

        self._plugins_view = plugins_view

    def _on_close_button_clicked(self, button):
        self.hide()

    def _on_plugins_treeview_cursor_changed(self, treeview):
        # Update info panel for the selected plugin
        row = treeview.get_cursor()[0][0]
        plugin = self._plugins[row]

        # Update plugin info
        self._name_label.set_markup('<big><b>%s</b></big>' %
                                    plugin['DISPLAY_NAME'])
        self._plugin_image.set_from_pixbuf(plugin['PIXBUF'])
        self._desc_label.set_text(plugin['DESCRIPTION'])
        self._author_label.set_text(plugin['AUTHOR'])
        self._copyright_label.set_text(plugin['COPYRIGHT'])

        self._url_linkbutton.set_label(plugin['URL'])
        self._url_linkbutton.set_uri(plugin['URL'])

        # Show/hide configure button
        is_active = treeview.get_model()[row][self.MODEL_ENABLED]
        mgr = mesk.plugin.get_manager()
        selected_plugin = self._plugins[row]
        plugin = mesk.plugin.get_manager().get_plugin(selected_plugin['NAME'])
        if is_active and plugin.is_configurable():
            self._config_button.set_sensitive(True)
        else:
            self._config_button.set_sensitive(False)

    def _on_plugin_toggled(self, cell, path, model):
        new_state = not model[path][0]
        path = int(path)
        plugin_name = self._plugins[path]['NAME']

        plugin_mgr = mesk.plugin.get_manager()
        state_str = ''
        try:
            if new_state:
                state_str = _('Plugin activation error')
                plugin_mgr.activate_plugin(plugin_name)
            else:
                state_str = _('Plugin deactivation error')
                plugin_mgr.deactivate_plugin(plugin_name)
            # Update model
            model[path][0] = new_state
        except Exception, ex:
            d = ErrorDialog(self.window)
            d.set_markup('<b>%s</b>' % state_str)
            d.format_secondary_text(str(ex))
            d.run()
            d.destroy()

    def _on_plugin_config_button_clicked(self, button):
        selected_row = self._plugins_view.get_cursor()[0][0]
        selected_plugin = self._plugins[selected_row]
        plugin = mesk.plugin.get_manager().get_plugin(selected_plugin['NAME'])

        self.config_widget = plugin.get_config_widget(self.window)
        self.plugin_config_container.pack_start(self.config_widget)
        self.plugin_config_container.show_all()
        self.close_button.set_sensitive(False)
        self._plugins_view.set_sensitive(False)
        self.plugins_notebook.set_current_page(1)
        self.config_plugin = plugin

    def _config_action(self, action):
        '''A helper for the config ok/cancel buttons'''
        action()

        self.plugin_config_container.remove(self.config_widget)
        self.config_widget.destroy()
        self.config_widget = None
        del self.config_widget
        self.config_plugin = None

        self.plugins_notebook.set_current_page(0)
        self.close_button.set_sensitive(True)
        self._plugins_view.set_sensitive(True)

    def _on_plugin_config_cancel_button_clicked(self, button):
        self._config_action(self.config_plugin.config_cancel)
    def _on_plugin_config_ok_button_clicked(self, button):
        self._config_action(self.config_plugin.config_ok)
